package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.model.validation.isNotNullAndNotBlank
import dev.fyloz.colorrecipesexplorer.repository.MaterialTypeRepository
import org.springframework.context.annotation.Profile
import org.springframework.stereotype.Service

interface MaterialTypeService :
  ExternalNamedModelService<MaterialType, MaterialTypeSaveDto, MaterialTypeUpdateDto, MaterialType, MaterialTypeRepository> {
    /** Checks if a material type with the given [prefix] exists. */
    fun existsByPrefix(prefix: String): Boolean

    /** Checks if the given [materialType] is used by one or more materials. */
    fun isUsedByMaterial(materialType: MaterialType): Boolean

    /** Gets all system material types. */
    fun getAllSystemTypes(): Collection<MaterialType>

    /** Gets all material types who are not a system type. */
    fun getAllNonSystemType(): Collection<MaterialType>
}

@Service
@Profile("!emergency")
class MaterialTypeServiceImpl(repository: MaterialTypeRepository, private val materialService: MaterialService) :
  AbstractExternalNamedModelService<MaterialType, MaterialTypeSaveDto, MaterialTypeUpdateDto, MaterialType, MaterialTypeRepository>(
    repository
  ), MaterialTypeService {
    override fun idNotFoundException(id: Long) = materialTypeIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = materialIdAlreadyExistsException(id)
    override fun nameNotFoundException(name: String) = materialTypeNameNotFoundException(name)
    override fun nameAlreadyExistsException(name: String) = materialTypeNameAlreadyExistsException(name)

    override fun MaterialType.toOutput() = this

    override fun existsByPrefix(prefix: String): Boolean = repository.existsByPrefix(prefix)
    override fun isUsedByMaterial(materialType: MaterialType): Boolean =
      materialService.existsByMaterialType(materialType)

    override fun getAllSystemTypes(): Collection<MaterialType> = repository.findAllBySystemTypeIs(true)
    override fun getAllNonSystemType(): Collection<MaterialType> = repository.findAllBySystemTypeIs(false)

    override fun save(entity: MaterialType): MaterialType {
        if (existsByPrefix(entity.prefix))
            throw materialTypePrefixAlreadyExistsException(entity.prefix)
        return super<AbstractExternalNamedModelService>.save(entity)
    }

    override fun update(entity: MaterialTypeUpdateDto): MaterialType {
        val persistedMaterialType by lazy { getById(entity.id) }

        return update(with(entity) {
            MaterialType(
              id = id,
              name = if (isNotNullAndNotBlank(name)) name else persistedMaterialType.name,
              prefix = if (isNotNullAndNotBlank(prefix)) prefix else persistedMaterialType.prefix,
              systemType = false
            )
        })
    }

    override fun update(entity: MaterialType): MaterialType {
        if (repository.existsByIdAndSystemTypeIsTrue(entity.id!!)) {
            throw cannotUpdateSystemMaterialTypeException(entity)
        }

        with(repository.findByPrefix(entity.prefix)) {
            if (this != null && id != entity.id)
                throw materialTypePrefixAlreadyExistsException(entity.prefix)
        }

        return super.update(entity)
    }

    override fun delete(entity: MaterialType) {
        if (repository.existsByIdAndSystemTypeIsTrue(entity.id!!)) {
            throw cannotDeleteSystemMaterialTypeException(entity)
        }

        if (!repository.canBeDeleted(entity.id)) throw cannotDeleteMaterialTypeException(entity)
        super.delete(entity)
    }
}
